package net.sf.saxon.functions;
import net.sf.saxon.CollectionURIResolver;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.trans.XPathException;

/**
 * Implement the fn:collection() function. The Saxon implementation loads an XML
 * document called the collection catalogue, which acts as an index of the collection.
 *
 * <p>The structure of this index is:
 * <pre>
 * &lt;collection&gt;
 *    &lt;doc href="doc1.xml"&gt;
 *    &lt;doc href="doc2.xml"&gt;
 *    &lt;doc href="doc3.xml"&gt;
 * &lt;/collection&gt;
 * </pre></p>
 *
 * <p>The document URIs are resolved relative to the base URI of the doc element
 * in the catalogue document.</p>
 */

public class Collection extends SystemFunction {

    private String expressionBaseURI = null;

    public void checkArguments(StaticContext env) throws XPathException {
        if (expressionBaseURI == null) {
            super.checkArguments(env);
            expressionBaseURI = env.getBaseURI();
        }
    }

    /**
    * preEvaluate: this method suppresses compile-time evaluation by doing nothing
    */

    public Expression preEvaluate(StaticContext env) {
        return this;
    }

    public SequenceIterator iterate(XPathContext context) throws XPathException {

        String href;

        // Zero-argument function:

        if (getNumberOfArguments() == 0) {
            href = null;
        } else {
            href = argument[0].evaluateItem(context).getStringValue();
        }

        CollectionURIResolver resolver = context.getConfiguration().getCollectionURIResolver();
        SequenceIterator iter = resolver.resolve(href, expressionBaseURI, context);

        if (iter == null) {
            return EmptyIterator.getInstance();
        } else {
            return iter;
        }
    }



    // TODO: provide control over error recovery (etc) through options in the catalog file.

}





//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
