package net.sf.saxon.functions;
import net.sf.saxon.Err;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.TypeHierarchy;
import net.sf.saxon.type.Type;
import net.sf.saxon.expr.*;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.sort.DescendingComparer;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.value.*;

import java.util.Comparator;

/**
* This class implements the min() and max() functions
*/

public class Minimax extends CollatingFunction {

    public static final int MIN = 2;
    public static final int MAX = 3;

    private boolean ignoreNaN = false;

    public void setIgnoreNaN(boolean ignore) {
        ignoreNaN = ignore;
    }

    /**
     * Static analysis: prevent sorting of the argument
     */

    public void checkArguments(StaticContext env) throws XPathException {
        // TODO: rewrite the expression if the static cardinality is (0-1), (0-0), or (1-1)
        super.checkArguments(env);
        Optimizer opt = env.getConfiguration().getOptimizer();
        argument[0] = ExpressionTool.unsorted(opt, argument[0], false);
    }

    /**
     * Perform optimisation of an expression and its subexpressions.
     * <p/>
     * <p>This method is called after all references to functions and variables have been resolved
     * to the declaration of the function or variable, and after all type checking has been done.</p>
     *
     * @param opt             the optimizer in use. This provides access to supporting functions; it also allows
     *                        different optimization strategies to be used in different circumstances.
     * @param env             the static context of the expression
     * @param contextItemType the static type of "." at the point where this expression is invoked.
     *                        The parameter is set to null if it is known statically that the context item will be undefined.
     *                        If the type of the context item is not known statically, the argument is set to
     *                        {@link net.sf.saxon.type.Type#ITEM_TYPE}
     * @return the original expression, rewritten if appropriate to optimize execution
     * @throws net.sf.saxon.trans.StaticError if an error is discovered during this phase
     *                                        (typically a type error)
     */

    public Expression optimize(Optimizer opt, StaticContext env, ItemType contextItemType) throws XPathException {
        Expression e = super.optimize(opt, env, contextItemType);
        if (e != this) {
            return e;
        }
        if (getNumberOfArguments() == 1) {
            // often happens after (A<B) is rewritten as (min(A) lt max(B))
            int card = argument[0].getCardinality();
            TypeHierarchy th = env.getNamePool().getTypeHierarchy();
            if (!Cardinality.allowsMany(card) && th.isSubType(argument[0].getItemType(th), Type.NUMBER_TYPE)) {
                return argument[0];
            }
        }
        return this;
    }

    /**
    * Evaluate the function
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        Comparator collator = getAtomicComparer(1, context);

        if (operation == MAX) {
            collator = new DescendingComparer(collator);
        }

        SequenceIterator iter = argument[0].iterate(context);

        // Get the first value in the sequence, ignoring any NaN values
        AtomicValue min;
        while (true) {
            min = (AtomicValue)iter.next();
            if (min == null) {
                return null;
            }
            if (min instanceof UntypedAtomicValue) {
                try {
                    min = new DoubleValue(Value.stringToNumber(min.getStringValueCS()));
                } catch (NumberFormatException e) {
                    dynamicError("Failure converting " +
                                                     Err.wrap(min.getStringValueCS()) +
                                                     " to a number", "FORG0001", context);
                }
            }
            if (min instanceof NumericValue && ((NumericValue)min).isNaN()) {
                // if there's a NaN in the sequence, return NaN, unless ignoreNaN is set
                if (ignoreNaN) {
                    continue;   // ignore the NaN and treat the next item as the first real one
                } else {
                    return min; // return NaN
                }
            } else {
                break;          // process the rest of the sequence
            }
        }

        while (true) {
            AtomicValue test = (AtomicValue)iter.next();
            if (test==null) break;
            AtomicValue test2 = test;
            if (test instanceof UntypedAtomicValue) {
                try {
                    test2 = new DoubleValue(Value.stringToNumber(test.getStringValueCS()));
                } catch (NumberFormatException e) {
                    dynamicError("Failure converting " +
                                                     Err.wrap(test.getStringValueCS()) +
                                                     " to a number", "FORG0001", context);
                }
            }
            if (test2 instanceof NumericValue && ((NumericValue)test2).isNaN()) {
                // if there's a NaN in the sequence, return NaN, unless ignoreNaN is set
                if (!ignoreNaN) {
                    return test2;
                }
            } else {
                try {
                    if (collator.compare(test2, min) < 0) {
                        min = test2;
                    }
                } catch (ClassCastException err) {
                    final TypeHierarchy th = context.getNamePool().getTypeHierarchy();
                    typeError("Cannot compare " + min.getItemType(th) +
                                       " with " + test2.getItemType(th), "FORG0006", context);
                    return null;
                }
            }
        }
        return min;
    }

}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
