package net.sf.saxon.functions;

import net.sf.saxon.*;
import net.sf.saxon.event.Builder;
import net.sf.saxon.event.Stripper;
import net.sf.saxon.event.PipelineConfiguration;
import net.sf.saxon.expr.MappingFunction;
import net.sf.saxon.expr.MappingIterator;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.*;
import net.sf.saxon.pattern.NodeKindTest;
import net.sf.saxon.trans.DynamicError;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.value.ObjectValue;
import net.sf.saxon.value.Whitespace;
import org.xml.sax.XMLReader;

import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.ErrorListener;
import javax.xml.transform.stream.StreamSource;
import java.io.File;
import java.io.FilenameFilter;
import java.net.URI;
import java.net.URISyntaxException;

/**
 * This class implements the default collection URI Resolver.
 * <p>
 * This supports two implementations of collections. If the URI supplied uses the "file:/" scheme, and the
 * file that is referenced is a directory, then the collection is the set of files in that directory. Query parameters
 * may be included in the URI:
 * <ul>
 * <li>recurse=yes|no controls whether the directory is scanned recursively; </li>
 * <li>strip=yes|no  determines whether whitespace text nodes are stripped from the selected documents; </li>
 * <li>val=strict|lax|preserve|strip determines whether schema validation is applied;</li>
 * <li>select=pattern determines which files in the directory are selected.</li>
 * <li>onerror=fail|warn|ignore determines the action taken if processing of a file fails</li>
 * <li>parser=qualified.class.name selects the parser (XMLReader) to be used to read the files</li>
 * </ul>
 * <p>
 * Otherwise, the resolver attempts to dereference the URI to obtain a catalog file. This is an XML file
 * containing a list of documents, in the format: </p>
 * <code><pre>
 * <collection>
 *   <doc href="doc1.xml"/>
 *   <doc href="doc2.xml"/>
 * </collection>
 * </pre></code>
 */

public class StandardCollectionURIResolver implements CollectionURIResolver, MappingFunction {

    /**
     * Resolve a URI.
     *
     * @param href The relative URI of the collection. This corresponds to the
     *             argument supplied to the collection() function. If the collection() function
     *             was called with no arguments (to get the "default collection") this argument
     *             will be null.
     * @param base The base URI that should be used. This is the base URI of the
     *             static context in which the call to collection() was made, typically the URI
     *             of the stylesheet or query module
     * @return an Iterator over the documents in the collection. The items returned
     *         by this iterator must implement the {@link net.sf.saxon.om.NodeInfo} interface.
     *         <p/>
     *         If the URI is not recognized, the method may either return an empty iterator,
     *         in which case no error is reported, or it may throw an exception, in which case
     *         the query or transformation fails. Returning null has the same effect as returning
     *         an empty iterator.
     */

    public SequenceIterator resolve(String href, String base, XPathContext context) throws XPathException {

        if (href == null) {
            // default collection. This returns empty, we previously threw an error.
            return null;
        }

        if (base == null) {
            base = StandardURIResolver.tryToExpand(base);
            if (base == null) {
                DynamicError err = new DynamicError(
                    "Cannot resolve relative URI: no base URI available");
                err.setXPathContext(context);
                throw err;
            }
        }

        URI resolvedURI;
        URIQueryParameters params = null;
        try {
            URI relative = new URI(href);
            String query = relative.getQuery();
            if (query != null) {
                params = new URIQueryParameters(query, context.getConfiguration());
                int q = href.indexOf('?');
                href = href.substring(0, q);
            }
            resolvedURI = new URI(base).resolve(href);
        } catch (URISyntaxException e) {
            DynamicError err = new DynamicError(
                    "Invalid URI " + Err.wrap(href) + " passed to collection() function");
            err.setXPathContext(context);
            throw err;
        }

        if ("file".equals(resolvedURI.getScheme())) {
            File file = new File(resolvedURI);
            if (!file.exists()) {
                DynamicError err = new DynamicError(
                        "The file or directory " + resolvedURI + " does not exist");
                err.setXPathContext(context);
                throw err;
            }
            if (file.isDirectory()) {
                return directoryContents(file, params, context);
            }
        }
        return catalogContents(resolvedURI, params, context);

    }

    private SequenceIterator directoryContents(File directory, URIQueryParameters params, XPathContext context) {

        FilenameFilter filter = null;

        if (params != null) {
            FilenameFilter f = params.getFilenameFilter();
            if (f != null) {
                filter = f;
            }
        }

        File[] files;
        if (filter == null) {
            files = directory.listFiles();
        } else {
            files = directory.listFiles(filter);
        }

        ObjectValue[] fileValues = new ObjectValue[files.length];
        for (int f=0; f<files.length; f++) {
            fileValues[f] = new ObjectValue(files[f]);
        }

        // If the URI requested suppression of errors, or that errors should be treated
        // as warnings, we set up a special ErrorListener to achieve this

        int onError = URIQueryParameters.ON_ERROR_FAIL;
        if (params != null && params.getOnError() != null) {
            onError = params.getOnError().intValue();
        }
        final Controller controller = context.getController();
        final PipelineConfiguration oldPipe = controller.makePipelineConfiguration();
        final PipelineConfiguration newPipe = new PipelineConfiguration(oldPipe);
        final ErrorListener oldErrorListener = controller.getErrorListener();
        if (onError == URIQueryParameters.ON_ERROR_IGNORE) {
            newPipe.setErrorListener(new ErrorListener() {
                public void warning(TransformerException exception) {}
                public void error(TransformerException exception) {}
                public void fatalError(TransformerException exception) {}
            });
        } else if (onError == URIQueryParameters.ON_ERROR_WARNING) {
            newPipe.setErrorListener(new ErrorListener() {
                public void warning(TransformerException exception) throws TransformerException {
                    oldErrorListener.warning(exception);
                }
                public void error(TransformerException exception) throws TransformerException {
                    oldErrorListener.warning(exception);
                    DynamicError supp = new DynamicError("The document will be excluded from the collection");
                    supp.setLocator(exception.getLocator());
                    oldErrorListener.warning(supp);
                }
                public void fatalError(TransformerException exception) throws TransformerException {
                    error(exception);
                }
            });
        }
        FileExpander expander = new FileExpander(params, newPipe);
        SequenceIterator base = new ArrayIterator(fileValues);
//        if (oldErrorListener != null) {
//            ClosingAction action = new ClosingAction() {
//                public void close(SequenceIterator base) {
//                    controller.setErrorListener(oldErrorListener);
//                }
//            };
//            base = new ClosingIterator(base, action);
//        }
        return new MappingIterator(base, expander, context);
    }

    private SequenceIterator catalogContents(URI catalogFile, URIQueryParameters params, XPathContext context)
    throws XPathException {

        DocumentInfo catalog =
                (DocumentInfo) Document.makeDoc(catalogFile.toString(), null, context, null);
        if (catalog==null) {
            // we failed to read the catalogue
            DynamicError err = new DynamicError("Failed to load collection catalogue " + catalogFile);
            err.setXPathContext(context);
            throw err;
        }

        // Now return an iterator over the documents that it refers to

        SequenceIterator iter =
                catalog.iterateAxis(Axis.CHILD, NodeKindTest.ELEMENT);
        NodeInfo top;
        while (true) {
            top = (NodeInfo)iter.next();
            if (top == null) break;
            if (!("collection".equals(top.getLocalPart()) &&
                    top.getURI().equals("") )) {
                DynamicError err = new DynamicError("collection catalogue must contain top-level element <collection>");
                err.setXPathContext(context);
                throw err;
            }
            break;
        }

        SequenceIterator documents =
                top.iterateAxis(Axis.CHILD, NodeKindTest.ELEMENT);

        return new MappingIterator(documents, this, context);
    }



    /**
     * Map from doc elements in the catalogue document to nodes
     * returned in the result
     * @param item A doc element in the catalogue document
     * @param context The dynamic evaluation context
     * @return the document or element referenced by the @href attribute of the doc
     * element in the catalogue
     * @throws net.sf.saxon.trans.XPathException if the document cannot be retrieved or parsed, unless
     * error recovery has been chosen.
     */

    public Object map(Item item, XPathContext context) throws XPathException {
        NodeInfo element = (NodeInfo)item;
        if (!("doc".equals(element.getLocalPart()) &&
                element.getURI().equals("") )) {
            DynamicError err = new DynamicError(
                    "children of <collection> element must be <doc> elements");
            err.setXPathContext(context);
            throw err;
        }
        String href = Navigator.getAttributeValue(element, "", "href");
        if (href==null) {
            DynamicError err = new DynamicError(
                    "\"<doc> element in catalogue has no @href attribute\"");
            err.setXPathContext(context);
            throw err;
        }

        NodeInfo target = Document.makeDoc(href, element.getBaseURI(), context, null);
        return target;
    }

    /**
     * Mapping function to process the files in a directory. This maps a sequence of external
     * objects representing files to a sequence of DocumentInfo nodes representing the parsed
     * contents of those files.
     */

    private static class FileExpander implements MappingFunction {

        private URIQueryParameters params;
        boolean recurse = false;
        int strip = Whitespace.UNSPECIFIED;
        int validation = Validation.STRIP;
        XMLReader parser = null;
        int onError = URIQueryParameters.ON_ERROR_FAIL;
        FilenameFilter filter = null;
        PipelineConfiguration pipe;

        public FileExpander(URIQueryParameters params, PipelineConfiguration pipe) {
            this.params = params;
            this.pipe = pipe;
            if (params != null) {
                FilenameFilter f = params.getFilenameFilter();
                if (f != null) {
                    filter = f;
                }
                Boolean r = params.getRecurse();
                if (r != null) {
                    recurse = r.booleanValue();
                }
                Integer v = params.getValidationMode();
                if (v != null) {
                    validation = v.intValue();
                }
                strip = params.getStripSpace();
                Integer e = params.getOnError();
                if (e != null) {
                    onError = e.intValue();
                }
                XMLReader p = params.getXMLReader();
                if (p != null) {
                    parser = p;
                }
            }

        }

        /**
         * Map one item to a sequence.
         *
         * @param item    The item to be mapped.
         *                If context is supplied, this must be the same as context.currentItem().
         * @param context The processing context. Some mapping functions use this because they require
         *                context information. Some mapping functions modify the context by maintaining the context item
         *                and position. In other cases, the context may be null.
         * @return either (a) a SequenceIterator over the sequence of items that the supplied input
         *         item maps to, or (b) an Item if it maps to a single item, or (c) null if it maps to an empty
         *         sequence.
         */

        public Object map(Item item, XPathContext context) throws XPathException {
            File file = (File)((ObjectValue)item).getObject();
            if (file.isDirectory()) {
                if (recurse) {
                   File[] files;
                    if (filter == null) {
                        files = file.listFiles();
                    } else {
                        files = file.listFiles(filter);
                    }

                    ObjectValue[] fileValues = new ObjectValue[files.length];
                    for (int f=0; f<files.length; f++) {
                        fileValues[f] = new ObjectValue(files[f]);
                    }

                    FileExpander expander = new FileExpander(params, pipe);
                    return new MappingIterator(new ArrayIterator(fileValues), expander, context);
                } else {
                    return null;
                }
            } else {
                try {
                    Source source = new StreamSource(file.toURI().toString());
                    if (validation != Validation.STRIP && validation != Validation.PRESERVE) {
                        source = AugmentedSource.makeAugmentedSource(source);
                        ((AugmentedSource)source).setSchemaValidationMode(validation);
                    }
                    if (parser != null) {
                        source = AugmentedSource.makeAugmentedSource(source);
                        ((AugmentedSource)source).setXMLReader(parser);
                    }

                    Stripper stripper = null;
                    if (params != null) {
                        int stripSpace = params.getStripSpace();
                        switch (strip) {
                            case Whitespace.ALL: {
                                stripper = AllElementStripper.getInstance();
                                stripper.setStripAll();
                                source = AugmentedSource.makeAugmentedSource(source);
                                ((AugmentedSource)source).addFilter(stripper);
                                break;
                            }
                            case Whitespace.IGNORABLE:
                            case Whitespace.NONE:
                                source = AugmentedSource.makeAugmentedSource(source);
                                ((AugmentedSource)source).setStripSpace(stripSpace);
                        }
                    }
                    NodeInfo contextNode = Builder.build(source, stripper, pipe);
                    return contextNode.getDocumentRoot();
                } catch (XPathException err) {
                    if (onError == URIQueryParameters.ON_ERROR_IGNORE) {
                        return null;
                    } else if (onError == URIQueryParameters.ON_ERROR_WARNING) {
                        try {
                            if (!err.hasBeenReported()) {
                                pipe.getErrorListener().warning(err);
                                DynamicError supp = new DynamicError("The document will be excluded from the collection");
                                supp.setLocator(err.getLocator());
                                pipe.getErrorListener().warning(supp);
                            }
                        } catch (TransformerException err2) {
                            //
                        }
                        return null;
                    } else {
                        throw err;
                    }
                }
            }
        }
    }




}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

