package net.sf.saxon.sort;

import net.sf.saxon.expr.*;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.type.AnyItemType;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.type.TypeHierarchy;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.ObjectValue;
import net.sf.saxon.value.Value;

import java.io.PrintStream;
import java.util.Iterator;

/**
 * A TupleSorter is an expression that sorts a stream of tuples. It is used
 * to implement XQuery FLWR expressions.
 */
public class TupleSorter extends ComputedExpression implements MappingFunction {

    private Expression base;
    private FixedSortKeyDefinition[] sortKeys;

        // Although this class uses the FixedSortKeyDefinition class to define the sort keys,
        // the actual sort key expression in the FixedSortKeyDefinition is not used. This is because
        // the sort key is instead defined as one of the members of the tuple delivered by the
        // TupleSorter. Therefore, the sort key expression is not managed as a child of this expression.

    public TupleSorter(Expression base, FixedSortKeyDefinition[] keys) {
        this.base = base;
        this.sortKeys = keys;
        adoptChildExpression(base);
    }

    public Expression simplify(StaticContext env) throws XPathException {
        base = base.simplify(env);
        return this;
    }

    public Expression typeCheck(StaticContext env, ItemType contextItemType) throws XPathException {
        base = base.typeCheck(env, contextItemType);
        return this;
    }

    public Expression optimize(Optimizer opt, StaticContext env, ItemType contextItemType) throws XPathException {
        base = base.optimize(opt, env, contextItemType);
        if (base instanceof EmptySequence) {
            return base;
        }
        return this;
    }

    public ItemType getItemType(TypeHierarchy th) {
        return AnyItemType.getInstance();
            // TODO: we can do better than this, but we need more information
    }

    public int computeCardinality() {
        return StaticProperty.ALLOWS_ZERO_OR_MORE;
    }

    public Iterator iterateSubExpressions() {
        return new MonoIterator(base);
    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        Expression exp = offer.accept(this);
        if (exp != null) {
            return exp;
        } else {
            base = base.promote(offer);
            return this;
        }
    }

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        SequenceIterator iter = new SortedTupleIterator(context, base.iterate(context), sortKeys);
        MappingIterator mapper = new MappingIterator(iter, this, context);
        return mapper;
    }

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        // so long as the sequence is homogeneous (all atomic values or all nodes), the EBV
        // of the sorted sequence is the same as the EBV of the base sequence. Only if it is
        // heterogeneous do we need to do the sort in order to calculate the EBV.
        final TypeHierarchy th = context.getNamePool().getTypeHierarchy();
        ItemType type = base.getItemType(th);
        if (type == Type.ITEM_TYPE) {
            return super.effectiveBooleanValue(context);
        } else {
            return base.effectiveBooleanValue(context);
        }
    }

    public void display(int level, NamePool pool, PrintStream out) {
        out.println(ExpressionTool.indent(level) + "TupleSorter");
        base.display(level+1, pool, out);
    }

    /**
     * Mapping function to map the wrapped objects returned by the SortedTupleIterator
     * into real items. This is done because each tuple may actually represent a sequence
     * of underlying values that share the same sort key.
     */

    public Object map(Item item, XPathContext context) throws XPathException {
        ObjectValue tuple = (ObjectValue)item;
        Object o = tuple.getObject();
        if (o == null) {
            return o;
        }
        if (o instanceof Item) {
            return o;
        }
        Value value = (Value)o;
        return value.iterate(context);
    }

}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none
//